/*
    pmacct (Promiscuous mode IP Accounting package)
    pmacct is Copyright (C) 2003-2006 by Paolo Lucente
*/

/*
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.


   NOTE: because the main idea is to make the pmacct classifier FULLY compatibile
   with the patterns of the L7-filter project, very little parts of this file have
   been grabbed - and carefully adapted in order to work into the new home :) - by
   the L7-filter project source code. They are marked as "l7code" and protected as
   follows:

   By Matthew Strait <quadong@users.sf.net>, Oct 2003.

   http://l7-filter.sf.net

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version
   2 of the License, or (at your option) any later version.
   http://www.gnu.org/licenses/gpl.txt
*/

#define __CLASSIFIER_C

#include "pmacct.h"
#include "pmacct-data.h"
#include "plugin_hooks.h"
#include "ip_flow.h"
#include "classifier.h"
#include "jhash.h"
#include "util.h"

u_int32_t class_trivial_hash_rnd = 140281;

void init_classifiers(char *path)
{
  char fname[MAX_FN_LEN];
  struct dirent **namelist;
  struct stat st;
  int entries = 0, n = 0, x = 0, ret;

  memset(&class, 0, sizeof(struct pkt_classifier)*MAX_CLASSIFIERS);
  entries = pm_scandir(path, &namelist, 0, pm_alphasort);
  if (entries > 0) {
    while (n < entries) {
      snprintf(fname, sizeof(fname), "%s/%s", path, namelist[n]->d_name);
      ret = stat(fname, &st);

      if (ret < 0) Log(LOG_ERR, "ERROR: Unable to stat(): '%s'. Skipped.\n", fname);
      else {
	if (S_ISREG(st.st_mode) && dot_pat(fname)) {
	  if (n > MAX_CLASSIFIERS) {
	    Log(LOG_DEBUG, "DEBUG: No more room for classifiers.\n");
	    break;
	  }

	  Log(LOG_DEBUG, "DEBUG: reading %s classifier.\n", fname);
	  ret = parse_pattern_file(fname, &class[x]);
	  if (ret) {
	    class[x].id = x+1; /* id are >= 1 */
	    x++;
	  }
	}
      }
      free(namelist[n]);
      n++;
    }
    free(namelist);
    Log(LOG_DEBUG, "DEBUG: %d classifiers successfully loaded.\n", x);
  }
  else {
    Log(LOG_ERR, "ERROR: Unable to open: '%s'\n", path);
    exit(1); 
  }
}

void evaluate_classifiers(struct packet_ptrs *pptrs, struct ip_flow_common *fp, unsigned int idx)
{
  int plen = (config.snaplen ? config.snaplen : DEFAULT_SNAPLEN);
  char payload[plen+1];
  int j = 0, ret;

  if (pptrs->new_flow) {
    init_class_accumulators(pptrs, fp, idx);
    search_conntrack(fp, pptrs, idx); 
  }

  /* Short circuit: a) if we have a class; b) if we have no more
     tentatives to classify the packet. Otherwise continue */
  if (fp->class[idx] || !fp->cst[idx].tentatives) {
    pptrs->class = fp->class[idx];
    handle_class_accumulators(pptrs, fp, idx);

    /* do we have an helper ? If yes, let's run it ! */
    if (fp->conntrack_helper) fp->conntrack_helper(fp->last[idx], pptrs);

    return;
  }

  /* We will pre-process the payload section of the snapshot */
  if (pptrs->payload_ptr) {
    int len = ((struct pcap_pkthdr *)pptrs->pkthdr)->caplen - (pptrs->payload_ptr - pptrs->packet_ptr), x = 0, y = 0;
 
    while (x < len && y < plen) {
      if (pptrs->payload_ptr[x] != '\0') {
        if (isascii(pptrs->payload_ptr[x])) payload[y] = tolower(pptrs->payload_ptr[x]);
	else payload[y] = pptrs->payload_ptr[x];
	y++;
      }
      x++;
    }
    payload[y] = '\0';

    while (class[j].pattern && j < MAX_CLASSIFIERS) {
      ret = regexec(class[j].pattern, payload);
      if (ret) {
        fp->class[0] = class[j].id;
        fp->class[1] = class[j].id;
        pptrs->class = class[j].id;
	handle_class_accumulators(pptrs, fp, idx);
	if (class[j].ct_helper) {
	  fp->conntrack_helper = class[j].ct_helper;
	  fp->conntrack_helper(fp->last[idx], pptrs); 
	}
	else fp->conntrack_helper = NULL;
        return;
      }
      j++;
    }
  }

  fp->class[idx] = FALSE;
  pptrs->class = FALSE;
  handle_class_accumulators(pptrs, fp, idx);
}

void init_class_accumulators(struct packet_ptrs *pptrs, struct ip_flow_common *fp, unsigned int idx)
{
  unsigned int reverse = idx ? 0 : 1;
  struct timeval now;

  memset(&pptrs->cst, 0, CSSz);
  memset(&fp->cst[idx], 0, CSSz);
  fp->cst[idx].tentatives = DEFAULT_TENTATIVES;
  fp->class[idx] = FALSE;

  memcpy(&fp->cst[idx].stamp, &fp->last[idx], sizeof(struct timeval));
  /* If the reciprocal of this flow a) is not expired and b) has a class
     then let's inherit it */
  now.tv_sec = fp->last[idx];
  now.tv_usec = 0;
  if (!is_expired_uni(&now, fp, reverse))
    fp->class[idx] = fp->class[reverse];
  else 
    fp->conntrack_helper = NULL; 
}

void handle_class_accumulators(struct packet_ptrs *pptrs, struct ip_flow_common *fp, unsigned int idx)
{
  struct my_iphdr *iphp = (struct my_iphdr *)pptrs->iph_ptr; 
#if defined ENABLE_IPV6
  struct ip6_hdr *ip6hp = (struct ip6_hdr *)pptrs->iph_ptr; 
#endif

  /* The flow doesn't have a class yet */
  if (!fp->class[idx]) { 
    /* We have more chances to classify the flow */ 
    if (fp->cst[idx].tentatives) {
      memset(&pptrs->cst, 0, CSSz);
      if (pptrs->l3_proto == ETHERTYPE_IP)
	fp->cst[idx].ba += ntohs(iphp->ip_len); 
#if defined ENABLE_IPV6
      else if (pptrs->l3_proto == ETHERTYPE_IPV6)
	fp->cst[idx].ba += (IP6HdrSz+ntohs(ip6hp->ip6_plen));
#endif
      if (pptrs->new_flow) fp->cst[idx].fa++;
      if (pptrs->pf) {
	fp->cst[idx].pa += pptrs->pf;
	pptrs->pf = 0;
      }
      fp->cst[idx].pa++;
      if (pptrs->payload_ptr) fp->cst[idx].tentatives--;
    }
    /* We finished tentatives, flow class is unknown */ 
    else {
      memset(&pptrs->cst, 0, CSSz);
      memset(&fp->cst[idx], 0, CSSz);
    }
  }
  else {
    memcpy(&pptrs->cst, &fp->cst[idx], CSSz);
    memset(&fp->cst[idx], 0, CSSz);
  } 
}

/* dot_pat(): checks that the supplied fname ends with the '.pat' suffix */
int dot_pat(char *fname)
{
  int len = strlen(fname);

  if (fname[len-4] == '.' && fname[len-3] == 'p' &&
      fname[len-2] == 'a' && fname[len-1] == 't') return 1; 

  return 0;
}

/* l7code */
static int hex2dec(char c)
{
  switch (c) {
  case '0':
  case '1':
  case '2':
  case '3':
  case '4':
  case '5':
  case '6':
  case '7':
  case '8':
  case '9':
    return c - '0';
  case 'a':
  case 'b':
  case 'c':
  case 'd':
  case 'e':
  case 'f':
    return c - 'a' + 10;
  case 'A':
  case 'B':
  case 'C':
  case 'D':
  case 'E':
  case 'F':
    return c - 'A' + 10;
  default:
    Log(LOG_ERR, "hex2dec(): bad value!\n");
    return 0;
  }
}

/* l7code: takes a string with \xHH escapes and returns one with the
   characters they stand for */
static char * pre_process(char * s)
{
  char *result = malloc(strlen(s) + 1);
  int sindex = 0, rindex = 0;

  while( sindex < strlen(s) ) {
    if( sindex + 3 < strlen(s) &&
        s[sindex] == '\\' && s[sindex+1] == 'x' &&
        isxdigit(s[sindex + 2]) && isxdigit(s[sindex + 3]) ) {
      /* carefully remember to call tolower here... */
      result[rindex] = tolower( hex2dec(s[sindex + 2])*16 + hex2dec(s[sindex + 3] ) );
      sindex += 3; /* 4 total */
    }
    else result[rindex] = tolower(s[sindex]);

    sindex++;
    rindex++;
  }
  result[rindex] = '\0';

  return result;
}

/* l7code */
int parse_pattern_file(char *fname, struct pkt_classifier *class)
{
  FILE *f;
  char line[MAX_PATTERN_LEN];
  int len = 0, linelen = 0, ret;

  enum { protocol, pattern, done } datatype = protocol;

  f = fopen(fname, "r");

  if (!f) return 0;

  while ( fgets(line, MAX_PATTERN_LEN, f) ) {
    linelen = strlen(line);
    if(linelen < 2 || line[0] == '#') continue;

    /* strip the pesky newline... */
    if (line[linelen - 1] == '\n') line[linelen - 1] = '\0';

    if (datatype == protocol) {
      if (linelen >= MAX_PROTOCOL_LEN) {
        Log(LOG_ERR, "ERROR: Protocol name in %s too long. A maximum of %d chars is allowed.\n", fname, MAX_PROTOCOL_LEN);
	return 0;
      }

      strncpy(class->protocol, line, MAX_PROTOCOL_LEN);
      datatype = pattern;
    }
    else if (datatype == pattern) {
      if (linelen >= MAX_PATTERN_LEN) {
        Log(LOG_ERR, "ERROR: Pattern in %s too long. A maximum of %d chars is allowed.\n", fname, MAX_PATTERN_LEN);
	return 0;
      }
      class->pattern = regcomp(pre_process(line), &linelen);
      if (!class->pattern) {
	Log(LOG_ERR, "ERROR: Failed compiling regular expression for protocol '%s'\n", class->protocol);
	return 0;
      }

      datatype = done;
      break;
    }
    else {
      Log(LOG_ERR, "ERROR: parse_pattern_file(): internal error\n");
      return 0;
    }
  }

  if (datatype != done) {
    Log(LOG_ERR, "ERROR: Failed to get all needed data from %s\n", fname);
    return 0;
  }

  fclose(f);

  link_conntrack_helper(class);

  return 1;
}

int pm_scandir(const char *dir, struct dirent ***namelist,
            int (*select)(const struct dirent *),
            int (*compar)(const struct dirent **, const struct dirent **))
{
  DIR *d;
  struct dirent *entry;
  register int i = 0;
  size_t entrysize;

  if ((d=opendir(dir)) == NULL) return(-1);

  *namelist = NULL;
  while ((entry=readdir(d)) != NULL)
  {
    if (select == NULL || (select != NULL && (*select)(entry)))
    {
      *namelist=(struct dirent **)realloc((void *)(*namelist), (size_t)((i+1)*sizeof(struct dirent *)));
      if (*namelist == NULL) return(-1);
      entrysize=sizeof(struct dirent)-sizeof(entry->d_name)+strlen(entry->d_name)+1;
      (*namelist)[i]=(struct dirent *)malloc(entrysize);
      if ((*namelist)[i] == NULL) return(-1);
      memcpy((*namelist)[i], entry, entrysize);
      i++;
    }
  }
  if (closedir(d)) return(-1);
  if (i == 0) return(-1);
  if (compar != NULL)
    qsort((void *)(*namelist), (size_t)i, sizeof(struct dirent *), compar);
    
  return(i);
}

int pm_alphasort(const struct dirent **a, const struct dirent **b)
{
  return(strcmp((*a)->d_name, (*b)->d_name));
}

void link_conntrack_helper(struct pkt_classifier *class)
{
  int index = 0;

  while (strcmp(conntrack_helper_list[index].protocol, "")) {
    if (!strcmp(class->protocol, conntrack_helper_list[index].protocol)) {
      class->ct_helper = conntrack_helper_list[index].ct_helper;
      break;
    }

    index++;
  }
}
